const Lang = imports.lang;

const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();

const FileUtils = Me.imports.fileUtils;
const PromiseUtils = Me.imports.promiseUtils;

const CACHE_DIR = `${GLib.get_user_cache_dir()}/${Me.metadata.uuid}`;
GLib.mkdir_with_parents(`${CACHE_DIR}`, 0o774);

const sortObj = (obj, criterion = (key) => key) => {
    return Object.keys(obj)
        .sort(criterion)
        .reduce(function (result, key) {
            result[key] = obj[key];
            return result;
        }, {});
};

var Index = new Lang.Class({
    Name: "IndexUtilsIndex",

    /**
     * Constructor for Index objects.
     *
     * An Index stores objects passed to it in JSON format.
     *
     * For searching purposes, specified keys' values are passed to a scorer function that returns
     * tokens mapped to scores. A reference to the object (sha256 hash) is stored for every token.
     *
     * Keys not specified at creation will be discarded.
     *
     * To search, a passed search query is split into tokens. This is handled by a tokenizer.
     *
     * Look at scorer.js and tokenizer.js for simple implementations.
     *
     * @param {string} indexId - A (file-path-safe) name/id for the index.
     * @param {string[]} keys - A list of keys that indexed objects will have.
     * @param {scorer} scorer - A function that returns tokens in an index object mapped to scores
     * @param {tokenizer} tokenizer - A function that returns the tokens in a query
     *
     * @return {Void}
     */
    _init: function (indexId, keys, scorer, tokenizer) {
        this.indexId = indexId;
        this.keys = keys;

        this.scorer = scorer;
        this.tokenizer = tokenizer;

        const indexedFile = Gio.File.new_for_path(`${CACHE_DIR}/${this.indexId}/indexed`);
        this.indexed = indexedFile.query_exists(null);

        this._checksum = GLib.Checksum.new(GLib.ChecksumType.SHA256);
        this._readyCallback = () => {};

        GLib.mkdir_with_parents(`${CACHE_DIR}/${this.indexId}/known`, 0o774);
        GLib.mkdir_with_parents(`${CACHE_DIR}/${this.indexId}/tokens`, 0o774);
    },

    /**
     * Destructor
     *
     * @return {Void}
     */
    destroy: function () {
        this._checksum.free(); // Not sure if I need to do this, better safe than sorry though
    },

    /**
     * Create an index entry for an object.
     *
     * @param {Object} indexObject - An object containing the keys specified at creation.
     *
     * @return {Promise<string>} - A promise that resolves with the entry's sha256 hash
     *                             once the index entry has been created.
     */
    createIndexEntry: function (indexObject) {
        const tokenScores = this.scorer(indexObject);

        const indexObjectHash = this._hashIndexObject(indexObject);

        return PromiseUtils.iteratePromises(Object.keys(tokenScores), (token) => {
            return FileUtils.readFileOr(`${CACHE_DIR}/${this.indexId}/tokens/${token}`, "{}").then(
                (fileContents) => {
                    const tokenScoreObj = JSON.parse(fileContents);

                    tokenScoreObj[indexObjectHash] = tokenScores[token];

                    return FileUtils.writeToFile(
                        `${CACHE_DIR}/${this.indexId}/tokens/${token}`,
                        JSON.stringify(sortObj(tokenScoreObj, (key) => -tokenScoreObj[key]))
                    );
                }
            );
        }).then(() => new Promise((resolve) => resolve(indexObjectHash)));
    },

    /**
     * Remove an index entry by sha256 hash.
     *
     * @param {string} hash - The hash identifying the entry.
     *
     * @return {Promise} A promise that resolves once the index entry has been deleted.
     */
    removeIndexEntry: function (hash) {
        return FileUtils.readFileOr(`${CACHE_DIR}/${this.indexId}/known/${hash}`)
            .then((fileContents) => {
                const indexObject = JSON.parse(fileContents);
                const tokenScores = this.scorer(indexObject);

                return PromiseUtils.iteratePromises(Object.keys(tokenScores), (token) => {
                    return FileUtils.readFileOr(
                        `${CACHE_DIR}/${this.indexId}/tokens/${token}`,
                        "{}"
                    ).then((tokenFileContents) => {
                        const tokenScoreObj = JSON.parse(tokenFileContents);

                        delete tokenScoreObj[hash];

                        return FileUtils.writeToFile(
                            `${CACHE_DIR}/${this.indexId}/tokens/${token}`,
                            JSON.stringify(sortObj(tokenScoreObj, (key) => tokenScoreObj[key]))
                        );
                    });
                });
            })
            .then(
                () =>
                    new Promise((resolve) => {
                        // Priority 0
                        Gio.File.new_for_path(
                            `${CACHE_DIR}/${this.indexId}/known/${hash}`
                        ).delete_async(0, null, (file, res) => resolve(file.delete_finish(res)));
                    })
            );
    },

    /**
     * Get whether the index has finished creating.
     *
     * This will be set to true once updateIndex finishes.
     * You can also manually mark an Index as ready by calling {@link markReady}
     *
     * @return {boolean} Whether the Index object is ready to be used.
     */
    isReady: function () {
        return this.indexed;
    },

    /**
     * Mark the Index as ready to use. Will call the Index's readyCallback, if set.
     *
     * @return {Void}
     */
    markReady: function () {
        this.indexed = true;

        const indexedFile = Gio.File.new_for_path(`${CACHE_DIR}/${this.indexId}/indexed`);
        if (!indexedFile.query_exists(null)) {
            indexedFile.create(Gio.FileCreateFlags.NONE, null).close(null);
        }

        if (this._readyCallback) this._readyCallback();
    },

    /**
     * Update the entire index to match an array of objects.
     *
     * This will remove all pre-existing index entries that are not
     * in the array and create new entries where necessary.
     *
     * @return {Promise} A promise that resolves once the Index is up-to-date.
     */
    updateIndex: function (allIndexObjects) {
        console.log(
            `gnome-fuzzy-app-search: updating index '${this.indexId}' with ${allIndexObjects.length} items`
        );

        return FileUtils.listDirectory(`${CACHE_DIR}/${this.indexId}/known`).then((files) => {
            const obsoleteIndexObjectHashes = files;

            return PromiseUtils.iteratePromises(
                allIndexObjects,
                (item, index) =>
                    new Promise((resolve) => {
                        const indexObjectHash = this._hashIndexObject(allIndexObjects[index]);

                        if (!obsoleteIndexObjectHashes.delete(indexObjectHash)) {
                            this.createIndexEntry(allIndexObjects[index])
                                .then(() => {
                                    return this._saveHash(indexObjectHash, allIndexObjects[index]);
                                })
                                .then(() => resolve());
                        } else {
                            resolve();
                        }
                    })
            ).then(() => {
                console.log(`gnome-fuzzy-app-search: Done adding to index '${this.indexId}'`);

                this.markReady();

                return PromiseUtils.iteratePromises(
                    Array.from(obsoleteIndexObjectHashes),
                    (obsoleteIndexObjectHash) => this.removeIndexEntry(obsoleteIndexObjectHash)
                );
            });
        });
    },

    /**
     * Get list of index objects by string query (split into words)
     *
     * Uses the tokenizer passed to the constructor
     *
     * @param  {string[]} query - The query, split into words
     * @param  {number} limit - How many matches to return at most
     *
     * @return {Promise<Object[]>} - Promise resolving to an array of index objects
     *                               that match the query, ranked from best to worst
     */
    find: function (query, limit = 10) {
        const indexObjectScores = {};
        const indexObjectResults = [];

        return PromiseUtils.iteratePromises(this.tokenizer(query), (token) => {
            return FileUtils.readFileOr(`${CACHE_DIR}/${this.indexId}/tokens/${token}`).then(
                (fileContents) => {
                    const tokenScoreObj = JSON.parse(fileContents);

                    Object.keys(tokenScoreObj)
                        .slice(0, limit - 1)
                        .forEach((hash) => {
                            indexObjectScores[hash] =
                                (indexObjectScores[hash] ?? 0) +
                                (token.length * tokenScoreObj[hash]) /
                                    Math.sqrt(Object.keys(tokenScoreObj).length);
                        });
                }
            );
        })
            .then(() => {
                return PromiseUtils.iteratePromises(
                    Object.keys(indexObjectScores)
                        .sort((a, b) => indexObjectScores[b] - indexObjectScores[a])
                        .slice(0, limit - 1),
                    (hash) =>
                        this.getObjectForHash(hash).then((indexObject) =>
                            indexObjectResults.unshift(indexObject)
                        )
                );
            })
            .then(() => indexObjectResults);
    },

    /**
     * Get the object that an entry refers to, by hash.
     *
     * @param  {string} hash - The sha256 hash to look up
     *
     * @return {Promise<Object>} - Promise resolving to the entry's object
     */
    getObjectForHash: function (hash) {
        return FileUtils.readFileOr(`${CACHE_DIR}/${this.indexId}/known/${hash}`).then(
            (fileContents) => JSON.parse(fileContents)
        );
    },

    /**
     * Function to be called once the index is populated
     * @callback readyCallback
     */

    /**
     * Call a function when updateIndex() finishes or markIndexed() is called.
     *
     * Overwrites any previous functions passed to this method.
     *
     * @param {readyCallback} callback - The function to call once the search is up and running
     * @return {Void}
     */
    setReadyCallback: function (callback = () => {}) {
        this._readyCallback = callback;
    },

    /**
     * Return an object containing only the keys passed at creation, in order passed at creation.
     *
     * @param {Object} indexObject - An object contining any keys in any order
     * @return {Object} An object containing only this.keys, in order of this.keys
     */
    _normalizeIndexObject: function (indexObject) {
        const strippedObject = {};

        this.keys.forEach((key) => {
            strippedObject[key] = indexObject[key];
        });

        return strippedObject;
    },

    /**
     * Normalize an index object, then turn it into JSON.
     *
     * @param {Object} indexObject - An object contining any keys in any order
     * @return {string} The object's normalized JSON representation.
     */
    _indexObjectToString: function (indexObject) {
        return JSON.stringify(this._normalizeIndexObject(indexObject));
    },

    /**
     * Normalize, JSONify and sha256-hash an object. Used for index entry "ID"s.
     *
     * @param {Object} indexObject - An object contining any keys in any order
     * @return {string} A hash unique to the index-relevant values in the object.
     */
    _hashIndexObject: function (indexObject) {
        this._checksum.reset();
        this._checksum.update(this._indexObjectToString(indexObject));
        const hash = this._checksum.get_string();
        this._checksum.reset();
        return hash;
    },

    /**
     * Save the (normalized) JSON of an index object under a hash
     *
     * @param {string} hash        - The hash to be used as entry "ID"
     * @param {Object} indexObject - The object to store
     * @return {Promise} A promise that resolves once the object has been stored.
     */
    _saveHash: function (hash, indexObject) {
        return FileUtils.writeToFile(
            `${CACHE_DIR}/${this.indexId}/known/${hash}`,
            this._indexObjectToString(indexObject)
        );
    },
});
