/**
 * A function to be called on each item of the list that may return a promise
 * @callback itemFunction
 * @param {Object} item - The current item
 * @param {number} index - The index of the item
 * @param {Object[]} array - The array the item came from, in case you forgot.
 * @return {Promise|Object} A promise that will be then'd, or something else.
 */

/**
 * Iterate over an array by calling a function on each element,
 * checking if the return is a Promise; if yes:
 *      call the function on the next element once the Promise resolves
 * if not:
 *      call the function on the next element immediately
 *
 * Returns a Promise.
 *
 * @param {Array} array - The array to iterate over
 * @param {itemFunction} itemFn - A function to call on each item of the array
 */
var iteratePromises = (array, itemFn = (item, index, array) => {}) => {
    return new Promise((resolve) => {
        let index = array.length;

        const loopBody = () => {
            index--;
            if (index < 0) return resolve();

            const promiseOrReturn = itemFn(array[index], index, array);

            if (
                (promiseOrReturn ?? null !== null) &&
                "then" in promiseOrReturn &&
                typeof promiseOrReturn.then === "function"
            ) {
                promiseOrReturn.then(() => loopBody());
            } else {
                loopBody();
            }
        };

        loopBody();
    });
};
